#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "computation_graph.h"

#define GRAPH_MAX_DEPTH 20
#define GRAPH_MAX_INPUTS 20

// Array to store input values for passing to the execute function of each node
static double exec_input_vals[GRAPH_MAX_INPUTS];

struct computation_graph *create_graph() {
    struct computation_graph *the_graph = malloc(sizeof(struct computation_graph));
    the_graph->nodes = malloc(sizeof(struct graph_node));
    the_graph->n_nodes = 0;
    the_graph->size = 1;
    return the_graph;
}

void reset_node(int node_id) {

}

int graph_set_source(struct computation_graph *graph,
                     int dest_cntl_id, int dest_input, int src_cntl_id, int src_output) {
    if (dest_cntl_id >= graph->n_nodes || src_cntl_id >= graph->n_nodes) {
        return -1;
    }
    struct graph_node *dest_node = &graph->nodes[dest_cntl_id];
    struct graph_node *src_node = &graph->nodes[src_cntl_id];
    if (dest_input >= dest_node->type->n_inputs || src_output >= src_node->type->n_outputs) {
        return -1;
    }

    dest_node->input_srcs[dest_input].controller_id = src_cntl_id;
    dest_node->input_srcs[dest_input].controller_output = src_output;
    reset_node(src_cntl_id);
    return 0;
}

int graph_add_node(struct computation_graph *graph,
                   const char* name,
                   const struct graph_node_type *type,
                   void *state) {
    assert(type->n_inputs <= GRAPH_MAX_INPUTS);
    int new_id = graph->n_nodes;
    if (new_id >= graph->size) {
        int new_capacity = graph->n_nodes == 0 ? 1 : graph->n_nodes * 2;
        struct graph_node *node_arr = realloc(graph->nodes, sizeof(struct graph_node) * new_capacity);
        if (!node_arr) {
            return -1;
        }
        graph->size = new_capacity;
        graph->nodes = node_arr;
    }
    struct graph_node *new_node = &graph->nodes[new_id];
    new_node->name = strdup(name);
    new_node->type = type;
    new_node->output_values = malloc(type->n_outputs * sizeof(double));
    new_node->param_values = malloc(type->n_params * sizeof(double));
    new_node->state = state;
    new_node->input_srcs = malloc(type->n_inputs * sizeof(struct input_type));
    int i;
    for (i = 0; i < type->n_inputs; i++) {
        new_node->input_srcs[i].controller_id = -1;
    }
    graph->n_nodes += 1;
    return new_id;
}

int graph_set_param_val(struct computation_graph *graph, int node_id, int param_id, double value) {
    if (node_id >= graph->n_nodes || param_id >= graph->nodes[node_id].type->n_params) {
        return -1;
    }
    graph->nodes[node_id].param_values[param_id] = value;
    return 0;
}

double graph_get_output(const struct computation_graph *graph, int node_id, int output_id) {
    if (node_id >= graph->n_nodes || output_id >= graph->nodes[node_id].type->n_outputs) {
        return -1;
    }
    return graph->nodes[node_id].output_values[output_id];
}

void graph_compute_node_rec(struct computation_graph *graph, int node_id, int depth) {
    if (depth >= GRAPH_MAX_DEPTH) {
        assert(1 == 0); // TODO :Xil_Assert false
        return;
    }
    if (node_id >= graph->n_nodes) {
        return;
    }
    // TODO: Fix self-reference loop
    struct graph_node *node = &graph->nodes[node_id];
    if (node->processed_state == PROCESSED) {
        return;
    }
    int input_id;
    for (input_id = 0; input_id < node->type->n_inputs; input_id++) {
        int src_cntl_id = node->input_srcs[input_id].controller_id;
        if (src_cntl_id != -1) {
            if (graph->nodes[src_cntl_id].processed_state == UNPROCESSED) {
                graph_compute_node_rec(graph, src_cntl_id, depth + 1);
            }
            int src_output_id = node->input_srcs[input_id].controller_output;
            exec_input_vals[input_id] = graph->nodes[src_cntl_id].output_values[src_output_id];
        }
    }
    (*node->type->execute)(node->state, node->param_values, exec_input_vals, node->output_values);
    node->processed_state = PROCESSED;
}

void graph_compute_node(struct computation_graph *graph, int node_id) {
    int i;
    for (i = 0; i < graph->n_nodes; i++) {
        graph->nodes[i].processed_state = UNPROCESSED;
    }
    graph_compute_node_rec(graph, node_id, 0);
}

int export_dot(const struct computation_graph* graph, FILE* of) {
    fprintf(of, "digraph G {\n"); // Header

    // Draw all the nodes and their inputs
    int i;
    for (i = 0; i < graph->n_nodes; i++) {
        struct graph_node *node = &graph->nodes[i];
        // Create node
        fprintf(of, "\"%s\"[label=\"", node->name);
        fprintf(of, "Name: %s\n", node->name); // Node name first
        int j;
        // Add parameters as labels
        for (j = 0; j < node->type->n_params; j++) {
            fprintf(of, "%s=%.3f\n", node->type->param_names[j], node->param_values[j]);
        }
        fprintf(of, "\"]\n"); // Close label bracket
        //
        for (j = 0; j < node->type->n_inputs; j++) {
            struct graph_node* src_node = &graph->nodes[node->input_srcs[j].controller_id];
            int output_id = node->input_srcs[j].controller_output;
            const char* output_name = src_node->type->output_names[output_id];
            fprintf(of, "\"%s\" -> \"%s\" [label=\"%s=%.3f\"]\n", src_node->name, node->name, output_name, src_node->output_values[output_id]);
        }
    }
    /*
    // Draw the links between nodes
    for (i = 0; i < graph->n_nodes; i++) {
        struct graph_node *node = &graph->nodes[i];
        int j;
        for (j = 0; j < node->n_inputs; j++) {
            // Input is connected if controller ID is not -1
            if (node->input_srcs[j].controller_id != -1) {
                struct graph_node* src_node = &graph->nodes[node->input_srcs[j].controller_id];
                int output_id = node->input_srcs[j].controller_output;
                const char* output_name = src_node->output_names[output_id];
                // Draws link from source controller to node input block. Also labels the value of the link
                fprintf(of, "\"%s\" -> \"%s[%s]\" [label=\"%s=%.3f\"]\n", src_node->name, node->name, node->input_names[j], output_name, src_node->output_values[output_id]);
            }
        }
    }
    */
    fprintf(of, "}"); // Close graph1
    return 0;
}